<?php

namespace Drupal\csp\EventSubscriber;

use Drupal\Core\Cache\CacheableResponseInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Render\AttachmentsInterface;
use Drupal\csp\Csp;
use Drupal\csp\CspEvents;
use Drupal\csp\Event\PolicyAlterEvent;
use Drupal\csp\Nonce;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Event subscriber to add CSP headers to responses.
 */
class ResponseCspSubscriber implements EventSubscriberInterface {

  /**
   * The Event Dispatcher service.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  private EventDispatcherInterface $eventDispatcher;

  /**
   * The Nonce service.
   *
   * @var \Drupal\csp\Nonce
   */
  private Nonce $nonce;

  /**
   * Constructs a new ResponseSubscriber object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   Config Factory service.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $eventDispatcher
   *   The Event Dispatcher Service.
   * @param \Drupal\csp\Nonce $nonce
   *   The nonce service.
   */
  public function __construct(
    protected ConfigFactoryInterface $configFactory,
    $eventDispatcher,
    $nonce,
  ) {
    $argOrder = FALSE;
    foreach (func_get_args() as $index => $arg) {
      if ($arg instanceof EventDispatcherInterface) {
        $this->eventDispatcher = $arg;
        $argOrder |= ($index !== 1);
      }
      elseif ($arg instanceof Nonce) {
        $this->nonce = $arg;
        $argOrder |= ($index !== 2);
      }
    }

    if ($argOrder) {
      // @phpcs:ignore Drupal.Semantics.FunctionTriggerError.TriggerErrorTextLayoutRelaxed
      @trigger_error("Parameter order has changed for 2.0.0.", E_USER_DEPRECATED);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events[KernelEvents::RESPONSE] = [
      // Nonce value needs to be added before settings are rendered to the page
      // by \Drupal\Core\EventSubscriber\HtmlResponseSubscriber.
      ['applyDrupalSettingsNonce', 1],
      // Policy needs to be generated after placeholder library info is bubbled
      // up and rendered to the page.
      ['onKernelResponse'],
    ];
    return $events;
  }

  /**
   * Add a nonce value to drupalSettings.
   *
   * The value is always added, but only available to scripts if a library
   * included on the response has csp/nonce or core/drupalSettings as a
   * dependency.
   *
   * @param \Symfony\Component\HttpKernel\Event\ResponseEvent $event
   *   The Response Event.
   */
  public function applyDrupalSettingsNonce(ResponseEvent $event): void {
    if (!$event->isMainRequest()) {
      return;
    }

    $response = $event->getResponse();
    if (!($response instanceof AttachmentsInterface)) {
      return;
    }

    $response->addAttachments([
      'drupalSettings' => [
        'csp' => [
          'nonce' => $this->nonce->getValue(),
        ],
      ],
    ]);
  }

  /**
   * Add Content-Security-Policy header to response.
   *
   * @param \Symfony\Component\HttpKernel\Event\ResponseEvent $event
   *   The Response event.
   */
  public function onKernelResponse(ResponseEvent $event): void {
    if (!$event->isMainRequest()) {
      return;
    }

    $cspConfig = $this->configFactory->get('csp.settings');

    $response = $event->getResponse();
    if ($response instanceof CacheableResponseInterface) {
      $response->getCacheableMetadata()
        ->addCacheableDependency($cspConfig);
    }

    foreach (['report-only', 'enforce'] as $policyType) {
      if (!$cspConfig->get($policyType . '.enable')) {
        continue;
      }

      $policy = new Csp();
      $policy->reportOnly($policyType == 'report-only');

      $this->eventDispatcher->dispatch(
        new PolicyAlterEvent($policy, $response),
        CspEvents::POLICY_ALTER
      );

      if (($headerValue = $policy->getHeaderValue())) {
        $response->headers->set($policy->getHeaderName(), $headerValue);
      }
    }
  }

}
